package com.vanheusden.pfa;

import java.util.*;

class Scene {
	/* private stuff */
	private Board board;
	private ChessObject [] white;
	private ChessObject [] black;
	private ChessObject kingWhite, kingBlack;
	private final List<Move> whiteMoves = new ArrayList<Move>(80);
	private final List<Move> blackMoves = new ArrayList<Move>(80);
	private boolean movesValidated = false, movesWhite = false, movesBlack = false;
	private PlayerColor movesValidatedColor;
	private final Move causingMove;
	private final ChessObject lastMoveObject;
	private final PlayerColor lastMoveColor;

	public void destructor() {
		board = null;

		white = null;
		black = null;

		whiteMoves.clear();
		blackMoves.clear();
	}

	boolean isValidated() {
		return movesValidated;
	}

	private void copyObjects(final ChessObject [] white, final ChessObject [] black) {
		assert white != null;
		assert black != null;
		assert this.white == null && this.black == null;

		final int nWhite = white.length;
		this.white = new ChessObject[nWhite];
		final int nBlack = black.length;
		this.black = new ChessObject[nBlack];

		for(int index=0; index<nWhite; index++) {
			final ChessObject cur = white[index].duplicate();

			if (cur instanceof ChessObjectKing)
				kingWhite = cur;

			this.white[index] = cur;
		}
		for(int index=0; index<nBlack; index++) {
			final ChessObject cur = black[index].duplicate();

			if (cur instanceof ChessObjectKing)
				kingBlack = cur;

			this.black[index] = cur;
		}
	}

	protected void killObject(final ChessObject victim, final ChessObject killer) {
		assert victim.getColor() != killer.getColor();

		/* if my color is WHITE, then his color is black so remove
		 * object from black 'array'
		 */
		if (killer.getColor() == PlayerColor.BLACK) {
			final int nIn = white.length;
			int nOut = 0;;
			final ChessObject [] newWhite = new ChessObject[nIn - 1];

			for(int index=0; index<nIn; index++) {
				if (white[index] != victim)
					newWhite[nOut++] = white[index];
			}

			white = newWhite;
		}
		else {
			final int nIn = black.length;
			int nOut = 0;
			final ChessObject [] newBlack = new ChessObject[nIn - 1];

			for(int index=0; index<nIn; index++) {
				if (black[index] != victim)
					newBlack[nOut++] = black[index];
			}

			black = newBlack;
		}
	}

	void replaceObject(ChessObject [] list, ChessObject old, ChessObject nw) {
		final int n = list.length;

		for(int index=0; index<n; index++) {
			if (list[index] == old) {
				list[index] = nw;
				break;
			}
		}
	}

	private void doMove(final Move move) {
		switch(move.getMoveType()) {
			case REGULAR:
				{
					final ChessObject object = board.getAt(move.getFX(), move.getFY());

					/* killing object? */
					final ChessObject dest_object = board.getAt(move.getTX(), move.getTY());
					if (dest_object != null)
						killObject(dest_object, object);

					/* do move object on board */
					object.doMove(move);
					board.moveObject(move);

					/* promote pawn? */
					if (object instanceof ChessObjectPawn && (object.getY() == 7 || object.getY() == 0)) {
						assert (object.getY() == 7 && object.getColor() == PlayerColor.WHITE) || (object.getY() == 0 && object.getColor() == PlayerColor.BLACK);

						final ChessObject newObject;

						if (move.getPromoteTo() != null)
							newObject = object.promote(move.getPromoteTo());
						else
							newObject = object.promote(ChessObjectType.QUEEN);

						board.replace(newObject);

						if (object.getColor() == PlayerColor.WHITE)
							replaceObject(white, object, newObject);
						else
							replaceObject(black, object, newObject);
					}
					else
					{
						assert move.getPromoteTo() == null;
					}

					break;
				}
			case CASTLING:	/* CASTLING */
				{
					final int y = move.getFY();

					assert move.getTX() == 2 || move.getTX() == 6;

					if (move.getTX() == 6) { // SHORT
						/* move king 4,y -> 6,y */
						board.moveObject(4, y, 6, y);
						board.getAt(6, y).doMove(6, y);

						/* move rook 7,y -> 5,y */
						board.moveObject(7, y, 5, y);
						board.getAt(5, y).doMove(5, y);
					}
					else {	// LONG (2)
						/* move king 4,y -> 2,y */
						board.moveObject(4, y, 2, y);
						board.getAt(2, y).doMove(2, y);

						/* move rook 0,y -> 3,y */
						board.moveObject(0, y, 3, y);
						board.getAt(3, y).doMove(3, y);
					}
					break;
				}
			case EN_PASSANT:	/* EN PASSANT */
				{
					final ChessObject object = board.getAt(move.getFX(), move.getFY());

					/* find location of the pawn that did that 2-forward move */
					final int delX = move.getTX();
					final int delY = object.getColor() == PlayerColor.WHITE ? 4 : 3;

					/* remove object from board and list */
					killObject(board.getAt(delX, delY), object);
					board.removeForEnPassant(delX, delY);

					/* now move the en-passant mover */
					object.doMove(move);
					board.moveObject(move);
					break;
				}
			default:
				assert false;
		}
	}

	public void createMoveList(final PlayerColor color) {
		Move lastMove = null;

                if (!movesWhite || !movesBlack) {
                        if (causingMove != null && lastMoveColor != color && lastMoveObject instanceof ChessObjectPawn) {
                                if (lastMoveObject.getNMoves() == 1 &&
                                                ((lastMoveColor == PlayerColor.BLACK && causingMove.getFY() == 6 && causingMove.getTY() == 4) ||
                                                 (lastMoveColor == PlayerColor.WHITE && causingMove.getFY() == 1 && causingMove.getTY() == 3)) &&
                                                board.getAt(causingMove.getTX(), lastMoveColor == PlayerColor.WHITE ? 2 : 5) == null) {  // or else it would be an normal move
                                        lastMove = causingMove;
                                }
                        }
                }
		// System.out.println("" + color + " " + lastMove);
		// System.out.println("" + causingMove + " " + lastMoveObject.getNMoves());

		if (!movesWhite && color == PlayerColor.WHITE) {
			final ChessObject kingWhite = getKing(PlayerColor.WHITE);

			for(int index=0; index<white.length; index++)
				white[index].getMoves(this, whiteMoves, kingWhite, lastMove);

			movesWhite = true;
		}

		if (!movesBlack && color == PlayerColor.BLACK) {
			final ChessObject kingBlack = getKing(PlayerColor.BLACK);

			for(int index=0; index<black.length; index++)
				black[index].getMoves(this, blackMoves, kingBlack, lastMove);

			movesBlack = true;
		}
	}

	/* public stuff */
	public Scene(final ChessObject [] white, final ChessObject [] black) {
		/* copy objects */
		copyObjects(white, black);

		/* initialize board */
		board = new Board(this.white, this.black);

		/* no move caused this board */
		causingMove = null;

		lastMoveObject = null;
		lastMoveColor = null;
	}

	public Scene(final ChessObject [] white, final ChessObject [] black, final Move move) {
		/* copy objects */
		copyObjects(white, black);

		/* initialize board */
		board = new Board(this.white, this.black);

		/* remember the move that "caused this situation" */
		causingMove = move;
		lastMoveObject = board.getAt(causingMove.getFX(), causingMove.getFY()); // FX/FY! move was not executed yet!
		lastMoveColor = lastMoveObject.getColor();

		/* move object */
		doMove(causingMove);
	}

	public Scene() {
		/* create whole new board */
		white = new ChessObject[16];
		int nWhite = 0;
		black = new ChessObject[16];
		int nBlack = 0;

		/* pawns */
		for(int x=0; x<8; x++) {
			white[nWhite++] = new ChessObjectPawn(x, 1, PlayerColor.WHITE);
			black[nBlack++] = new ChessObjectPawn(x, 6, PlayerColor.BLACK);
		}

		/* rooks */
		white[nWhite++] = new ChessObjectRook(0, 0, PlayerColor.WHITE);
		black[nBlack++] = new ChessObjectRook(0, 7, PlayerColor.BLACK);
		white[nWhite++] = new ChessObjectRook(7, 0, PlayerColor.WHITE);
		black[nBlack++] = new ChessObjectRook(7, 7, PlayerColor.BLACK);

		/* knight */
		white[nWhite++] = new ChessObjectKnight(1, 0, PlayerColor.WHITE);
		black[nBlack++] = new ChessObjectKnight(1, 7, PlayerColor.BLACK);
		white[nWhite++] = new ChessObjectKnight(6, 0, PlayerColor.WHITE);
		black[nBlack++] = new ChessObjectKnight(6, 7, PlayerColor.BLACK);

		/* knight */
		white[nWhite++] = new ChessObjectBishop(2, 0, PlayerColor.WHITE);
		black[nBlack++] = new ChessObjectBishop(2, 7, PlayerColor.BLACK);
		white[nWhite++] = new ChessObjectBishop(5, 0, PlayerColor.WHITE);
		black[nBlack++] = new ChessObjectBishop(5, 7, PlayerColor.BLACK);

		/* queen */
		white[nWhite++] = new ChessObjectQueen(3, 0, PlayerColor.WHITE);
		black[nBlack++] = new ChessObjectQueen(3, 7, PlayerColor.BLACK);

		/* king */
		kingWhite = new ChessObjectKing(4, 0, PlayerColor.WHITE);
		white[nWhite++] = kingWhite;
		kingBlack = new ChessObjectKing(4, 7, PlayerColor.BLACK);
		black[nBlack++] = kingBlack;

		/* initialize board */
		board = new Board(white, black);

		/* no move caused this board */
		causingMove = null;

		lastMoveObject = null;
		lastMoveColor = null;
	}

	public Scene Move(final Move move) {
		return new Scene(white, black, move);
	}

	public boolean isKingUnderAttack(final PlayerColor kingColor) {
		final PlayerColor opponentColor = Statics.opponentColor(kingColor);

		createMoveList(opponentColor);

		return getKing(kingColor).getNCatchers() != 0;
	}

	public boolean isOpponentPawnAt(final int x, final int y, final PlayerColor hisColor) {
		final ChessObject currentObject = board.getAt(x, y);

		return currentObject != null && currentObject instanceof ChessObjectPawn && currentObject.getColor() == hisColor;
	}

	public void validateMoves(final PlayerColor color) {
		movesValidatedColor = color;

		/* get the reference to the list of moves for the color selected */
		final List<Move> checkMoves = getMoveList(color);

		final PlayerColor opponentColor = Statics.opponentColor(color);
		final boolean kingUnderAttack = isKingUnderAttack(color);

		/* check every move 'color' can do */
		for(int moveIndex=0; moveIndex<checkMoves.size();) {
			boolean remove = false;

			final Move currentMove = checkMoves.get(moveIndex);

			ChessObject moveObject = board.getAt(currentMove.getFX(), currentMove.getFY());

			boolean needThoroughValidityCheck = kingUnderAttack ||
				moveObject instanceof ChessObjectKing ||
				currentMove.getMoveType() != MoveType.REGULAR ||
				moveObject.getKingPinned() == true;

			/* calculate moves of opponent AFTER we did the move we're checking */
			if (needThoroughValidityCheck && currentMove.getScene() == null) {
				Scene afterMove = Move(currentMove);
				currentMove.setScene(afterMove);
			}

			/* is my king under attack if I did this move? */
			if (needThoroughValidityCheck && currentMove.getScene().isKingUnderAttack(color) == true) {
				remove = true;
			}
			else if (currentMove.getMoveType() == MoveType.CASTLING) {
				assert (currentMove.getFY() == 0 && color == PlayerColor.WHITE) || (currentMove.getFY() == 7 && color == PlayerColor.BLACK);
				assert currentMove.getFY() == currentMove.getTY();

				if (kingUnderAttack)
					remove = true;
				else {
					// LONG castling = 2
					int checkField = currentMove.getTX() == 2 ? 3 : 5;
					int pawnLine = color == PlayerColor.WHITE ? 1 : 6;

					if (isFieldUnderAttack(opponentColor, checkField, currentMove.getFY()) || 
							isOpponentPawnAt(checkField - 1, pawnLine, opponentColor) ||
							isOpponentPawnAt(checkField + 1, pawnLine, opponentColor)) {
						remove = true;
					}
				}
			}

			if (remove) {
				/* move is invalid */
				checkMoves.remove(moveIndex);
			}
			else {
				moveIndex++;
			}
		}

		movesValidated = true;
	}

	public List<Move> getMoveList(final PlayerColor color) {
		if (color == PlayerColor.WHITE) {
			createMoveList(PlayerColor.WHITE);

			return whiteMoves;
		}

		createMoveList(PlayerColor.BLACK);

		return blackMoves;
	}

	public Board getBoard() {
		return board;
	}

	public ChessObject getObjectAt(final int x, final int y) {
		return board.getAt(x, y);
	}

	public ChessObject getKing(final PlayerColor color) {
		final ChessObject result = color == PlayerColor.WHITE ? kingWhite : kingBlack;
		assert result != null;
		return result;
	}

	public ChessObject [] getObjects(final PlayerColor color) {
		return color == PlayerColor.WHITE ? white : black;
	}

	public boolean isFieldUnderAttack(final PlayerColor opponentColor, final int x, final int y) {
		final List<Move> moveList = getMoveList(opponentColor);
		final int n = moveList.size();

		for(int index=0; index<n; index++) {
			final Move currentMove = moveList.get(index);

			if (currentMove.getTX() == x && currentMove.getTY() == y)
				return true;
		}

		return false;
	}

	public boolean isMoveInMovelist(Move move, PlayerColor color) {
		final List<Move> list = getMoveList(color);
		final int n = list.size();

		for(int index=0; index<n; index++) {
			final Move currentMove = list.get(index);

			if (currentMove.compareTo(move, true, false) == true)
				return true;
		}

		return false;
	}

	public boolean isCheckMate(PlayerColor side) {
		return getMoveList(side).size() == 0 && isKingUnderAttack(side) == true;
	}

	public boolean isStaleMate(PlayerColor side) {
		return getMoveList(side).size() == 0 && isKingUnderAttack(side) == false;
	}

	public boolean isInsufficientMaterialDraw() {
		int [] white = countMaterial(getObjects(PlayerColor.WHITE));
		int [] black = countMaterial(getObjects(PlayerColor.BLACK));

		// K+B v K, K+N v K, K v K

		if (white[5] + black[5] + // pawn
			white[2] + black[2] + // rook
			white[1] + black[1] != 0) // queen
			return false;

		if (white[3] + black[3] +	// knight
			white[4] + black[4] > 1) // bishop
			return false;

		return true;
	}

	int [] countMaterial(final ChessObject [] objs) {
		int [] material = new int[6];

		for(final ChessObject cur : objs)
			material[cur.getType().getSVI()]++;

		return material;
	}
}
