package com.vanheusden.pfa;

import java.io.*;
import java.util.concurrent.*;
import java.util.concurrent.atomic.*;
import java.util.concurrent.locks.*;

public class TranspositionTable implements Runnable {
	final Semaphore [] locks = new Semaphore[256];
	long [] lockUsage = new long[256];
	//
	int nTableElements;
	TTElement [] elements;
	//
	final String transTabFile;
	//
	final IO io;
	int ttfIOInterval = 2000;
	final Semaphore storeTPTDataLock = new Semaphore(1, false);
	//
	final Zobrist z;

	public TranspositionTable(String transTabFile, int nElements, IO io) throws IOException {
		long prevTs = System.currentTimeMillis();

		this.io = io;

		z = new Zobrist();

		for(int index=0; index<locks.length; index++)
			locks[index] = new Semaphore(1, false);

		this.transTabFile = transTabFile;

		nTableElements = nElements;

		if (nTableElements > 0) {
			elements = new TTElement[nTableElements];

			if (transTabFile != null && (new File(transTabFile)).exists()) {
				FileInputStream fin = new FileInputStream(transTabFile);
				BufferedInputStream bis = new BufferedInputStream(fin);
				DataInputStream in = new DataInputStream(bis);

				int inNElements = in.readInt();

				if (io != null)
					io.progressCallback("Load transposition table: " + inNElements + " elements");

				int nLoaded = 0;

				for(int loadIndex=0; loadIndex<inNElements; loadIndex++) {
					long curTs = System.currentTimeMillis();
					long diffTs = curTs - prevTs;
					if (io != null && diffTs >= ttfIOInterval) {
						io.progressCallbackDebug("Load TPT: " + String.format("%.2f%%", (loadIndex * 100.0) / nTableElements));

						prevTs = curTs;
					}

					long hash = -1;

					try {
						hash = in.readLong();
					}
					catch(EOFException ee) {
						break;
					}

					int index = calcArrayIndex(hash);
					short score = -1;
					Move m = null;
					Flag flag = null;
					byte depth = in.readByte();

					nLoaded++;

					score = in.readShort();

					byte b1 = in.readByte();
					byte b2 = in.readByte();
					int types = in.readByte();
					byte flagIn = in.readByte();

					int fx = (b1 >> 4) & 0x0f;
					int fy = b1 & 0x0f;
					int tx = (b2 >> 4) & 0x0f;
					int ty = b2 & 0x0f;
					assert(fx < 8);
					assert(fy < 8);
					assert(tx < 8);
					assert(ty < 8);

					MoveType mt = MoveType.getMT((byte)((types >> 4) & 0x0f));

					byte cotv = (byte)(types & 0x0f);
					ChessObjectType cot = cotv != 15 ? ChessObjectType.getCOT(cotv) : null;

					if (cot == null)
						m = new Move(fx, fy, tx, ty, mt);
					else
						m = new Move(fx, fy, tx, ty, cot);

					flag = Flag.getFlag(flagIn);

					elements[index] = new TTElement(hash, score, depth, m, flag, 0);
				}

				in.close();
				bis.close();
				fin.close();

				io.progressCallback("Transpositiontable usage: " + String.format("%06.2f", (nLoaded * 100.0) / inNElements) + "%");
			}
		}

		// test();
	}

	public Zobrist getZobrist() {
		return z;
	}

	public void destructor() throws IOException {
		if (io != null && transTabFile != null) {
			io.progressCallback("Dump transpositiontable to " + transTabFile);

			storeTPTData();
		}

		// wait untill any storing-thread has stopped
		storeTPTDataLock.acquireUninterruptibly();
		storeTPTDataLock.release();

		elements = null;
	}

	public void run() {
		try {
			Thread.currentThread().setPriority(Thread.MIN_PRIORITY);

			storeTPTData();
		}
		catch(IOException ie) {
			if (io != null)
				io.exception(ie);
		}
	}

	public int getNUsed() {
		int cnt = 0;

		for(int index=0; index<nTableElements; index++) {
			if (elements[index] != null)
				cnt++;
		}

		return cnt;
	}

	public void storeTPTData() throws IOException {
		if (nTableElements <= 0)
			return;

		long prevTs = System.currentTimeMillis();

		// only try storing the data if a filename is set
		// and we're not already storing the data
		if (transTabFile != null && storeTPTDataLock.tryAcquire()) {
			FileOutputStream fout = new FileOutputStream(transTabFile);
			BufferedOutputStream buffOut = new BufferedOutputStream(fout);
			DataOutputStream out = new DataOutputStream(buffOut);

			out.writeInt(nTableElements);

			int nFilled = 0;
			for(int index=0; index<nTableElements; index++) {
				long curTs = System.currentTimeMillis();
				long diffTs = curTs - prevTs;

				if (io != null && diffTs >= ttfIOInterval) {
					io.progressCallbackDebug("Store TPT: " + String.format("%.2f%%", (index * 100.0) / nTableElements));
					prevTs = curTs;

					buffOut.flush();
				}

				TTElement e = elements[index];

				Move m = e != null ? e.getMove() : null;
				if (m != null) {
					nFilled++;

					out.writeLong(e.getHash());

					out.writeByte(e.getDepth());

					out.writeShort(e.getEvalRaw());

					assert(m.getFX() < 8);
					assert(m.getFY() < 8);
					assert(m.getTX() < 8);
					assert(m.getTY() < 8);
					out.writeByte((m.getFX() << 4) | m.getFY());
					out.writeByte((m.getTX() << 4) | m.getTY());

					ChessObjectType pt = m.getPromoteTo();
					int ptv = pt != null ? pt.getSVI() : 15;

					out.writeByte((m.getMoveType().getType() << 4) | ptv);

					out.writeByte(e.getFlag().getV());
				}
			}

			out.close();
			buffOut.close();
			fout.close();

			storeTPTDataLock.release();

			io.progressCallback("Transpositiontable usage: " + String.format("%06.2f", (nFilled * 100.0) / nTableElements) + "%");
		}
	}

	public double [] getLockingStats() {
		double avg = 0, sd = 0;

		for(int i=0; i<locks.length; i++) {
			lock(i);

			avg += lockUsage[i];
			sd += lockUsage[i] * lockUsage[i];

			unlock(i);
		}

		double outAvg = avg / locks.length;
		double outSd = Math.sqrt(sd / locks.length - Math.pow(outAvg, 2.0));

		return new double [] { outAvg, outSd };
	}

	protected int lock(final long hash) {
		// limiting to 8 bits: hopefully this is big enough
		// for systems with even 4096 processing units (as that
		// would result in 8 threads per hash-slot)
		final int lockIndex = (int)(hash & 255);

		locks[lockIndex].acquireUninterruptibly();
		lockUsage[lockIndex]++;

		return lockIndex;
	}

	protected void unlock(final int lockIndex) {
		locks[lockIndex].release();
	}

	int calcArrayIndex(final long hash) {
		return (int)(hash % nTableElements);
	}

	public void put(final long hash, final Move m, final int ttScore, final byte depth, final Flag flag, final int ply) {
		if (nTableElements <= 0)
			return;

		int arrayIndex = calcArrayIndex(hash);

		int lockId = lock(hash);

		if (elements[arrayIndex] != null)
			elements[arrayIndex].set(hash, ttScore, depth, m, flag, ply);
		else
			elements[arrayIndex] = new TTElement(hash, ttScore, depth, m, flag, ply);

		unlock(lockId);
	}

	public TTElement get(final long hash) {
		if (nTableElements <= 0)
			return null;

		int arrayIndex = calcArrayIndex(hash);

		int lockId = lock(hash);

		TTElement result = elements[arrayIndex];

		unlock(lockId);

		if (result != null && result.getHash() == hash)
			return result;

		return null;
	}
}
