package com.vanheusden.pfa;

import java.security.SecureRandom;

class Zobrist {
	long [][] table = new long[64][32];
	long isWhite;

	final static long mask = 0x7fffffffffffffffl;

	public Zobrist(final long [][] table, final long isWhite) {
		this.table = table;
		this.isWhite = isWhite;
	}

	public Zobrist() {
		final SecureRandom sr = new SecureRandom();

		// the transposition table code cannot use negative values
		for(int pos=0; pos<64; pos++) {
			for(int type=0; type<32; type++)
				table[pos][type] = sr.nextLong() & mask;
		}

		isWhite = sr.nextLong() & mask;

		for(int pos=0; pos<64; pos++) {
			for(int type=0; type<32; type++) {
				assert isWhite != table[pos][type];

				for(int pos2=0; pos2<64; pos2++) {
					for(int type2=0; type2<32; type2++) {
						if (pos == pos2 && type == type2)
							continue;

						assert table[pos][type] != table[pos2][type2];
					}
				}
			}
		}
	}

	long calc(final ChessObject cur, final int offset) {
		int type = cur.getType().getSVI();

		if (cur instanceof ChessObjectPawn || cur instanceof ChessObjectRook || cur instanceof ChessObjectKing)
			type |= cur.isFirstMove() ? 8 : 0;

		type |= cur.getColor() == PlayerColor.BLACK ? 16 : 0;

		return table[offset][type];
	}

	long calc(final ChessObject [] white, final ChessObject [] black, final PlayerColor side) {
		long hash = 0;

		if (side == PlayerColor.WHITE)
			hash = isWhite;

		for(int index=0; index<white.length; index++) {
			ChessObject cur = white[index];

			hash ^= calc(cur, cur.getY() * 8 + cur.getX());
		}

		for(int index=0; index<black.length; index++) {
			ChessObject cur = black[index];

			hash ^= calc(cur, cur.getY() * 8 + cur.getX());
		}

		return hash;
	}

	long calc(long boardHashIn, Board before, Scene after, Move m, PlayerColor side) {
		if (m.getMoveType() == MoveType.REGULAR) {
			long newBoardHash = boardHashIn;

			// remove old position
			newBoardHash ^= calc(before.getAt(m.getFX(), m.getFY()), m.getFY() * 8 + m.getFX());

			final int destOffset = m.getTY() * 8 + m.getTX();

			// remove object on destination field (if any)
			final ChessObject dest = before.getAt(m.getTX(), m.getTY());

			if (dest != null)
				newBoardHash ^= calc(dest, destOffset);

			newBoardHash ^= isWhite;

			// put at new position
			Board afterBoard = after.getBoard();

			newBoardHash ^= calc(afterBoard.getAt(m.getTX(), m.getTY()), destOffset);

			return newBoardHash;
		}

		return calc(after.getObjects(PlayerColor.WHITE), after.getObjects(PlayerColor.BLACK), side);
	}

	long calc(final Board b, final PlayerColor side) {
		long hash = 0;

		if (side == PlayerColor.WHITE)
			hash ^= isWhite;

		for(int y=0; y<8; y++) {
			final int y8 = y * 8;

			for(int x=0; x<8; x++) {
				final ChessObject cur = b.getAt(x, y);

				if (cur != null)
					hash ^= calc(cur, y8 + x);
			}
		}

		return hash;
	}
}
